;;;
;;; groups.scm --- group definitions
;;;

(define group-operation (make-generic-operator 1
                                               (lambda (g)
                                                 (error "Unknown group operation for group" g))))
(define group-elements (make-generic-operator 1
                                              (lambda (g)
                                                (error "Unknown group elements for group" g))))
; takes group and element
(define is-element? (make-threaded-solver #f 2))

(define group-order (make-threaded-solver #f 1))

;; useful later
(define any? (lambda (x) #t))

;; for a speedup sometimes
(define (tag-inverse! g inverse-operation)
  (set-tag! g 'inverse inverse-operation))

;;;
;;; Straightforward finite groups
;;;

(define (make-group elements operation)
  (vector 'group elements operation))

(define (is-group-object? g)
  (and (vector? g) (eq? (vector-ref g 0) 'group)))

(defhandler group-operation
  (lambda (g)
    (vector-ref g 2))
  is-group-object?)

(defhandler group-elements
  (lambda (g)
    (vector-ref g 1))
  is-group-object?)

;; and some things which would work for any group
(defsolver is-element?
  (lambda (g e)
    (and (member e (group-elements g)) #t)))

(defsolver group-order
  (lambda (g)
    (length (group-elements g))))

(define (make-cyclic-group order)
  (let ((g (make-group (let lp ((i 0))
                         (if (< i order)
                             (cons i (lp (+ i 1)))
                             '()))
                       (lambda (a b)
                         (modulo (+ a b) order)))))
    (tag-inverse! g (lambda (a)
                      (modulo (- order a) order)))
    g))

;; for permutation groups:
(define (apply-perm p1 p2)
  (if (not (pair? p2))
      '()
      (cons (list-ref p1 (car p2)) (apply-perm p1 (cdr p2)))))
(define (perm-inverse p1)
  (let lp ((p p1)
           (i 0))
    (if (null? p)
        '()
        (cons (- (length p1) (length (member i p1))) (lp (cdr p) (+ i 1))))))

(define (make-symmetric-group n)
  (define (prepend e lis)
    (if (null? lis)
        (list (list e))
        (map (lambda (l)
               (cons e l))
             lis)))
  (let ((g (make-group
            (let lp ((elts (list-tabulate n (lambda (i) i))))
              (apply append (map (lambda (e)
                                   (prepend e
                                            (lp (delete e elts))))
                                 elts)))
            apply-perm)))
    (tag-inverse! g perm-inverse)
    g))

;;; lazy generator groups

(define (make-generator-group generators operation)
  (vector 'generator-group generators operation))
(define (is-generator-group-object? g)
  (and (vector? g) (eq? 'generator-group (vector-ref g 0))))
(define (generator-group-generators g)
  (vector-ref g 1))

(defhandler group-operation
  (lambda (g)
    (vector-ref g 2))
  is-generator-group-object?)

(defhandler group-elements
  (lambda (g)
    (let ((new-g (generate-group (generator-group-generators g) (group-operation g))))
      (vector-set! g 0 'group)
      (vector-set! g 1 (group-elements new-g))
      (vector-ref g 1)))
  is-generator-group-object?)

(defsolver is-element?
  (lambda (g e)
    (if (not (is-generator-group-object? g))
        unknown
        (let ((generators (make-equal-hash-table)))
          (let ((inv (and (has-tag? g 'inverse) (group-inverse-operation g))))
            (for-each (lambda (e)
                        (hash-table/put! generators e #t)
                        (if (has-tag? g 'inverse)
                            (hash-table/put! generators (inv e) #t)))
                      (generator-group-generators g)))
          (and (solve-to-element g e (hash-table/key-list generators)
                                 (car (generator-group-generators g))) #t))))) ;; silly... doesn't always work

;;; Group checking stuff

(define group-closed?
  (make-threaded-solver #f 1))

(defsolver group-closed?
  (lambda (g)
    (for-all? (group-elements g)
              (lambda (a)
                (for-all? (group-elements g)
                          (lambda (b)
                            (is-element? g ((group-operation g) a b))))))))

(defsolver group-closed?
  (lambda (g)
    (if (cyclic? g)
        #t
        unknown)))

(define group-associativity?
  (make-threaded-solver #f 1))

(defsolver group-associativity?
  (lambda (g)
    (for-all? (group-elements g)
              (lambda (a)
                (for-all? (group-elements g)
                          (lambda (b)
                            (for-all? (group-elements g)
                                      (lambda (c)
                                        (equal? ((group-operation g) a ((group-operation g) b c))
                                                ((group-operation g) ((group-operation g) a b) c))))))))))

(defsolver group-associativity?
  (lambda (g)
    (if (cyclic? g)
        #t
        unknown)))

(define group-identity
  (make-threaded-solver (lambda (ret g)
                          (tag-identity! g ret))
                        1))
(define no-identity (list 'no-identity))

(define (tag-identity! g ident)
  (set-tag! g 'identity ident))

(defsolver group-identity
  (lambda (g)
    (let lpe ((to-check (group-elements g)))
      (let lpa ((elt (group-elements g)))
        (cond ((null? to-check) no-identity)
              ((null? elt) (car to-check))
              ((and (equal? ((group-operation g) (car elt) (car to-check)) (car elt))
                    (equal? ((group-operation g) (car to-check) (car elt)) (car elt)))
               (lpa (cdr elt)))
              (else (lpe (cdr to-check))))))))

(define cyclic?
  (make-threaded-solver #f 1))

(defsolver cyclic?
  (lambda (g)
    (let lp ((elts (group-elements g))
             (table (make-equal-hash-table)))
      (cond ((null? elts) #f)
            ((= (group-order g)
                (let lp2 ((a ((group-operation g) (car elts) (car elts)))
                          (len (group-order g)))
                  (if (and (is-element? g a) (> len 0))
                      (begin
                        (hash-table/put! table a #t)
                        (lp2 ((group-operation g) (car elts) a) (- len 1)))
                      (hash-table/count table))))
             #t)
            (else (lp (cdr elts) (make-equal-hash-table)))))))

(define group-invertible?
  (make-threaded-solver #f 1))

(defsolver group-invertible?
  (lambda (g)
    (for-all? (group-elements g)
              (lambda (a)
                (not (eq? no-inverse (group-inverse g a)))))))

(define group-inverse
  (make-threaded-solver #f 2))
(define no-inverse (list 'no-inverse))

(defsolver group-inverse
  (lambda (g a)
    ((group-inverse-operation g) a))) ;; faster if we have a saved group inverse

(defsolver group-inverse
  (lambda (g a)
    (let lp ((elts (group-elements g)))
      (cond ((null? elts) no-inverse)
            ((and (equal? (group-identity g)
                          ((group-operation g) a (car elts)))
                  (equal? (group-identity g)
                          ((group-operation g) (car elts) a)))
             (car elts))
            (else (lp (cdr elts)))))))

(define group?
  (make-threaded-solver #f 1))

(defsolver group?
  (lambda (g)
    (if (cyclic? g)
        #t
        unknown)))

(defsolver group?
  (lambda (g)
    (and (group-closed? g)
         (group-associativity? g)
         (not (eq? no-inverse (group-identity g)))
         (group-invertible? g))))

(define group-inverse-operation
  (make-threaded-solver #f 1))

(defsolver group-inverse-operation
  (lambda (g)
    (get-tag g 'inverse unknown)))

(defsolver group-inverse-operation
  (lambda (g)
    (let ((inverses
           (map (lambda (a)
                  (group-inverse g a))
                (group-elements g))))
      (lambda (a)
        (let lp ((elts (group-elements g))
                 (inv inverses))
          (cond ((null? elts) no-inverse)
                ((equal? a (car elts)) (car inv))
                (else (lp (cdr elts) (cdr inv)))))))))

(define commutative?
  (make-threaded-solver #f 1))

(defsolver commutative?
 (lambda (g)
   (for-all? (group-elements g)
             (lambda (a)
               (for-all? (group-elements g)
                         (lambda (b)
                           (equal? ((group-operation g) a b)
                                   ((group-operation g) b a))))))))

(define abelian?
  (make-threaded-solver #f 1))

(defsolver abelian?
  (lambda (g)
    (and (group? g) (commutative? g))))

;;; examples

#|

;; (define g (make-generator-group '(1) +))
;; (is-element? g 4) => #t
;; (tag-inverse! g (lambda (e) (- e)))
;; (group-inverse g 4) => -4
;; (is-element? -4) -> #t

|#